#!/bin/bash

# This example shows what happens if we try to access the same key twice.
#
# This one's a little trickier than you may think at first glance.
#
# To begin with, we start a server with key generation rate and storage
# capacity set so high that availability of key material should not cause
# a problem.
#
# Well, the effect of the client's min_rate QoS requirement on the
# key generation bandwidth allocation should not be neglected.  The
# requested rate is 32 byte/s.  The server allocates 1.5x the requested
# rate => 48 byte/s, much less than the overall rate of 800 byte/s.
#
# Here is what the first part of the test does.
# - The client tries to immediately access key with index 2 and fails.
#   (Status: insufficient key material available)
# - After a 3s pause, it tries again.  Better luck this time.
#   (Status: success)
# - Again, it tries to access the key with index 2 and gets a slap on
#   the wrist.
#   (Status: key already spent)
# - Now it tries to access key with index 1.  That works, as expected.
#   (Status: success)
# - Again it tries to access key with index 1.  It fails.  But not
#   with "key already spent".  Instead, we get an "index out of bounds"
#   as the ringbuffer has moved its front index after consuming key 1
#   (it could not yet move the front index after key 2 was consumed as
#   key was was still available).
#   Life's full of surprises, innit?
#
# In the second part of the test, we substantially increase the
# min_bps requirement (and add a small 1s delay prior to accessing
# the first key, for good measure).  Now we see a sequence of
# - Status: success
# - Status: key already spent
# - Status: key already spent
# - Status: success
# - Status: index out of bounds

# First part of the test

test_dir=`dirname ${BASH_SOURCE[0]}`
test_name="trying_to_access_the_same_key_twice"

# valgrind helpers (active if ${RUN_TESTS_UNDER_VALGRIND}=1)
. ${test_dir}/../../.valgrind

# Remove lingering semaphores, if any.
env QKD_SERVER_FORCE_SEM_UNLINK=1 ./server_example 1001 A

# The actual test starts here.
timeout 8s ${memcheck1} ./server_example 1001 A ${test_dir}/high_key_rate.cfg &
sleep 2
${memcheck2} ./client_example ${test_dir}/empty.cfg 0 2 3 2 0 2 0 1 0 1 | grep Status > ${test_name}.A 2>&1
sleep 6

# Second part of the test: higher min_rate
sleep 1
timeout 8s ${memcheck3} ./server_example 1001 A ${test_dir}/high_key_rate.cfg &
sleep 2
${memcheck4} ./client_example ${test_dir}/higher_min_bps.cfg 1 2 3 2 0 2 0 1 0 1 | grep Status > ${test_name}.A2 2>&1
sleep 6

# NOP unless ${RUN_TESTS_UNDER_VALGRIND}=1
print_valgrind_error_summary

# Remove lingering semaphores, if any.
env QKD_SERVER_FORCE_SEM_UNLINK=1 ./server_example 1001 A

# Check results of both parts of the test
diff ${test_name}.A ${test_dir}/reflog && diff ${test_name}.A2 ${test_dir}/reflog2
