#ifndef ETSI_GS_QKD_004_CFG_PARSER_H
#define ETSI_GS_QKD_004_CFG_PARSER_H

/**
 * @file cfg_parser.h
 *
 * @section DESCRIPTION
 *
 * A very simple configuration file parser.
 * - Empty lines are ignored.
 * - Lines starting with '#' are considered comments and are
 *   ignored too.
 * - Other lines contain key-value pairs.
 * - Format: KEYWORD = VALUE
 * - The set of known keywords is specified by the key-value-table
 *   (i.e., a kv_pair_t array) given as argument to cfg_from_file().
 * - Unknown keywords are ignored after generating an error message.
 */

/**
 * A key-value pair.
 * Ownership of val is with the pair.
 * The key is supposed to be a compile-time constant that does not need
 * to be deallocated.
 */
typedef struct cfg_kv_pair {
    const char *key;
    char *val;
} cfg_kv_pair_t;

/**
 * A "configuration table" is an array of key-value pairs that has
 * - Non-NULL keys stored in its first N entries,
 * - without any duplicate keys, and
 * - a NULL key as an end-marker in its final (N+1st) entry.
 */
typedef cfg_kv_pair_t *cfg_table_t;

/**
 * free() all non-NULL value entries in the table.
 */
void cfg_table_dtor(cfg_table_t table);

/**
 * Access configuration table entry as string.
 * Returns 0 on success.
 * Returns 1 if the value is NULL.
 * Returns -1 if the no matching key was found.
 */
int cfg_get_string(cfg_table_t table, const char *key, const char **val);

/**
 * Access configuration table entry as long.
 * Returns 0 on success.
 * Returns 1 if the value in the table is NULL.
 * Returns 2 if strtol() failed.
 * Returns -1 if the no matching keyword was found.
 */
int cfg_get_long(cfg_table_t table, const char *key, long *val);

/**
 * Load a configuration from file.
 * Returns 0 on success.
 * Returns -1 if the configuration file could not be opened
 * (or if NULL was given as filename).
 */
int cfg_from_file(cfg_table_t table, const char *filename);

#endif // ETSI_GS_QKD_004_CFG_PARSER_H
