#ifndef ETSI_GS_QKD_004_SSL_SOCKET_H
#define ETSI_GS_QKD_004_SSL_SOCKET_H

#include <openssl/ssl.h>
#include "qkd/utils.h"

#ifdef __cplusplus__
extern "C" {
#endif

typedef qkd_bytes_t socket_data_t;
#define socket_data_dtor qkd_bytes_dtor

/**
 * Read num_bytes_to_read (ATTN: must be non-zero) from SSL connection.
 * Caller is responsible for deallocating returned data.
 * The .size field of the return value indicates the number of bytes
 * read, which, in case of transmission errors, may be smaller than
 * num_bytes to read.  The caller is responsible for checking .size.
 *
 * @param ssl               (in) SSL connection; must not be NULL
 * @param num_bytes_to_read (in) Number of bytes to be read
 * @return Data read; the caller is responsible for checking .size
 */
socket_data_t socket_ssl_recv(SSL *ssl, size_t num_bytes_to_read);

/**
 * Send data via the given SSL connection.
 * @param ssl (in) SSL connection; must not be NULL
 * @param data_bytes (in) Data to be written
 * @return 0 in case of success; a non-zero value if not all data
 *         could be sent successfully
 */
int socket_ssl_send(SSL *ssl, socket_data_t data_bytes);

typedef enum role { ROLE_CLIENT, ROLE_SERVER } role_t;

/**
 * Create and initialize an SSL connection.  Under the hood, a socket
 * connection to the host is created as well as an SSL context object.
 * For the sake of simplicity, we do not re-use this context object for
 * multiple SSL connections; it is uniquely associated with the SSL
 * connection created here.
 * The caller of socket_client_init(.) takes ownership of the returned
 * SSL object and its underlying context obect (cf. SSL_get_SSL_CTX(.))
 * and has to free both when they are no longer used.
 * If no socket could be created, the calling process is terminated
 * with an error code.
 *
 * @param (in) Target host
 * @param (in) Target port
 * @param (in) Client certificate
 * @param (in) Private key of client
 * @param (in) CA certificate(s) (one PEM file can contain multiple
 * certificates)
 * @return Pointer to SSL object created
 */
SSL *socket_client_init(const char *host, uint16_t port, const char *cert_pem,
                        const char *key_pem, const char *peer_ca_pem,
                        int min_proto_version, int max_proto_version);

int socket_create_server(const char *host, uint16_t port);

typedef enum server_status {
    SERVER_FINISHED,
    SERVER_LISTENING
} server_status_t;

void socket_server_init(const char *host, uint16_t port, const char *cert_pem,
                        const char *key_pem, const char *ca_pem,
                        int min_proto_version, int max_proto_version,
                        unsigned max_num_sessions,
                        void (*servlet)(SSL *ssl_socket,
                                        void *server_context_data,
                                        server_status_t *server_status),
                        void *server_context_data);

#ifdef __cplusplus__
}
#endif

#endif // ETSI_GS_QKD_004_SSL_SOCKET_H
