#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "qkd/cfg_parser.h"
#include "qkd/models.h"
#include "qkd/ssl_socket_client.h"
#include "qkd/utils.h"

#include <openssl/ssl.h>

static long *wait_times_and_indices = NULL;
static unsigned num_iterations = 0;

#define CFG_NUM_KEYWORDS 5
/**
 * Table of (optional) configuration file entries.
 * - port_number = the server's port number
 * - min_bps = minimum key rate in bits per second
 * - ttf = time to fetch in seconds
 * - client_uuid_string = string representation of client UUID
 *   (Must be in a format suited for uuid_parse(.).)
 * - capacity = QoS Capacity requirement (for OPEN_CONNECT)
 */
static cfg_kv_pair_t cfg_table[CFG_NUM_KEYWORDS + 1] = {
    {"port_number", NULL}, {"min_bps", NULL},
    {"ttf", NULL},         {"client_uuid_string", NULL},
    {"capacity", NULL},    {NULL, NULL} // the end-marker
};

// Configuration parameters and default values
static uint16_t port_number = QKD_004_DEFAULT_PORTNUMBER;
static uint32_t key_chunk_size = 32;
static uint32_t min_bps = 8 * 32;     // 32 byte per second
static uint32_t max_bps = 2 * 8 * 32; // 2 * min_bps
static uint32_t capacity = 0; // 0 => accept default capacity offered by server
static uint32_t timeout = 0;
static uint32_t ttf = 0;
static const char *client_uuid_string = "00000000-0000-0000-0000-000000000001";

// Exit codes
#define EXIT_CODE_INCORRECT_USAGE 1
#define EXIT_CODE_OUT_OF_MEMORY 2
#define EXIT_CODE_COULD_NOT_LOAD_CONFIG_FILE 3
#define EXIT_CODE_INCORRECT_CONFIG_PARAM 3
#define EXIT_CODE_OPEN_CONNECT_FAILED 5

static void clean_exit(int status) {
    free(wait_times_and_indices);
    exit(status);
}

static void print_usage_and_exit(const char *program_name) {
    assert(program_name != NULL);
    fprintf(stderr,
            "Usage: %s [<config-file>] [<wait-time> <index>] ...\n\n"
            "  <config-file> is the name of a (potentially empty) "
            "configuration file.\n\n"
            "  <config-file> is followed by an optional sequence of 2-tupels "
            "of <wait-time> and <index>.\n\n"
            "  If no such sequence is specified, \"1 0\" is assumed, i.e., a "
            "single key with index=0 is \n"
            "  requested after an initial wait time of 1 second.\n\n",
            program_name);
    clean_exit(EXIT_CODE_INCORRECT_USAGE);
}

int main(int argc, char *argv[]) {
    uint16_t port_number = QKD_004_DEFAULT_PORTNUMBER;
    char *invalid_chars = NULL;

    // Check command-line arguments
    if (argc > 1 && argc % 2 != 0) {
        print_usage_and_exit(argv[0]);
    }

    num_iterations = (argc > 2) ? (argc - 2) / 2 : 1;
    wait_times_and_indices =
        malloc(2 * num_iterations * sizeof(wait_times_and_indices[0]));
    if (wait_times_and_indices == NULL) {
        fprintf(stderr, "ERROR: could not allocate memory\n");
        clean_exit(EXIT_CODE_OUT_OF_MEMORY);
    }

    if (argc > 1) {
        printf("Reading configuration file %s ...\n", argv[1]);
        if (cfg_from_file(cfg_table, argv[1]) != 0) {
            fprintf(stderr, "Loading the configuration file failed\n");
            clean_exit(EXIT_CODE_COULD_NOT_LOAD_CONFIG_FILE);
        } else {
            long val;
            const char *stringval;

            int status = cfg_get_long(cfg_table, "port_number", &val);
            if (status == 0) {
                if (val < 0 || val > 65535) {
                    fprintf(stderr,
                            "Invalid port ID specified in configuration file "
                            "(valid range: 0..65535), skipping entry ...\n");
                    clean_exit(EXIT_CODE_INCORRECT_CONFIG_PARAM);
                } else {
                    port_number = val;
                    printf("Port number set to %d\n", port_number);
                }
            } else if (status == 2) {
                fprintf(stderr, "Invalid port_number entry.\n");
                clean_exit(EXIT_CODE_INCORRECT_CONFIG_PARAM);
            }

            status = cfg_get_long(cfg_table, "min_bps", &val);
            if (status == 0) {
                if (val < 0) {
                    fprintf(stderr, "Negative min_bps entry.\n");
                    clean_exit(EXIT_CODE_INCORRECT_CONFIG_PARAM);
                } else {
                    min_bps = val;
                    printf("min_bps set to %u\n", min_bps);
                }
            } else if (status == 2) {
                fprintf(stderr, "Invalid min_bps entry.\n");
                clean_exit(EXIT_CODE_INCORRECT_CONFIG_PARAM);
            }

            status = cfg_get_long(cfg_table, "ttf", &val);
            if (status == 0) {
                if (val < 0) {
                    fprintf(stderr, "Negative ttf entry.\n");
                    clean_exit(EXIT_CODE_INCORRECT_CONFIG_PARAM);
                } else {
                    ttf = val;
                    printf("TTF set to %u\n", ttf);
                }
            } else if (status == 2) {
                fprintf(stderr, "Invalid ttf entry.\n");
                clean_exit(EXIT_CODE_INCORRECT_CONFIG_PARAM);
            }

            status =
                cfg_get_string(cfg_table, "client_uuid_string", &stringval);
            if (status == 0) {
                uuid_t tmp;
                if (uuid_parse(stringval, tmp) != 0) {
                    fprintf(stderr, "Invalid client_uuid_string entry.\n");
                    clean_exit(EXIT_CODE_INCORRECT_CONFIG_PARAM);
                } else {
                    client_uuid_string = stringval;
                }
            }

            status = cfg_get_long(cfg_table, "capacity", &val);
            if (status == 0) {
                if (val < 0) {
                    fprintf(stderr, "Negative capacity entry.\n");
                    clean_exit(EXIT_CODE_INCORRECT_CONFIG_PARAM);
                } else {
                    capacity = val;
                    printf("Capacity (QoS parameter) set to %u\n", capacity);
                }
            } else if (status == 2) {
                fprintf(stderr, "Invalid capacity entry.\n");
                clean_exit(EXIT_CODE_INCORRECT_CONFIG_PARAM);
            }
        }
    }

    if (argc <= 2) {
        wait_times_and_indices[0] = 1; // 1 second initial wait time
        wait_times_and_indices[1] = 0; // key request w/ zero index then
    } else {
        for (int i = 2; i < argc; i++) {
            long val = wait_times_and_indices[i - 2] =
                strtol(argv[i], &invalid_chars, 10);
            if (!strcmp(argv[i], "") || val < 0 || *invalid_chars != 0) {
                print_usage_and_exit(argv[0]);
            }
        }
    }

    qkd_register_client_interface_routines();

    /* Activate/deactivate certificate availability */
    char *cert_pem = "../ssl/127.0.0.1.pem", *key_pem = "../ssl/127.0.0.1.key";
    // char *cert_pem = NULL, *key_pem = NULL;

    /* Activate/deactivate peer cert verification */
    // char *peer_ca_pem = "ssl/root_CA.pem";
    char *peer_ca_pem = NULL;
    char *host = "127.0.0.1";

    // Supported TLS protocol version range
    int min_proto_version = TLS1_3_VERSION;
    int max_proto_version = TLS1_3_VERSION;

    qkd_client_context_data_t cfg = {
        host,        port_number,       cert_pem,         key_pem,
        peer_ca_pem, min_proto_version, max_proto_version};

    // --- ECHO ---
    // for (int i=0; i<1000; i++) {
    char *msg = "Hello, world!";
    qkd_echo_request_t echo_request = {strlen(msg) + 1, msg};
    qkd_echo_response_t echo_response =
        qkd_echo_response_struct(echo_request, &cfg);
    qkd_fprintf_bytes(stdout, echo_response, 0);
    qkd_bytes_dtor(&echo_response);
    // }

    // --- OPEN_CONNECT ---
#if YOU_WANT_TO_SEE_SOME_ERROR_CODES
    char *destination_kme_id = "kme.evolutionq.de:12345";
    char *source_sae_id = "qkd://app1@aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa";
    char *destination_sae_id =
        "qkd://app2@bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb";
#else
    // NB:
    // - KME IDs and SAE IDs are URIs.
    // - We consider a URI to be a "NULL URI" if its first byte is zero.
    //   =>  char* null_URI = ""; // Good! Empty string => first byte is zero
    //       char* bad_URI  = NULL; // Says "NULL" ... but, sadly, this is a
    //                              // protocol violation (malformed data).

    // Our mock-up server has no connection to a QKDN controller.
    // => It cannot handle non-NULL destination KME IDs.
    char *destination_kme_id = "";
    // It is not connected to an application discovery service
    // either. => Destination SAE ID must be NULL.
    char *source_sae_id = "";
    // To avoid a protocol error, we must give a NULL source SAE ID if
    // the destination SAE ID is NULL.
    char *destination_sae_id = "";
#endif
    qkd_qos_t qos = {key_chunk_size, max_bps, min_bps, capacity,
                     timeout,        ttf,     "",      ""};
    qkd_open_connect_request_t oc_request = {
        destination_kme_id, source_sae_id, destination_sae_id, qos, {}, {}};
    uuid_parse(client_uuid_string, oc_request.key_stream_id);

    // qkd_open_connect() or qkd_open_connect_struct()
    qkd_open_connect_response_t oc_response =
        qkd_open_connect_struct(oc_request, &cfg);
    if (oc_response.status != qkd_status_code_successful) {
        fflush(stdout);
        fprintf(stderr, "Error: OPEN_CONNECT failed (");
        qkd_fprintf_status(stderr, oc_response.status);
        fprintf(stderr, ")\n");
        qkd_open_connect_response_dtor(&oc_response);
        clean_exit(EXIT_CODE_OPEN_CONNECT_FAILED);
    }

    // --- GET_KEY ---
    qkd_get_key_request_t gk_request = {
        {}, 0, (qkd_metadata_t){0, "[\"device_info\"]"}};
    uuid_copy(gk_request.key_stream_id, oc_response.key_stream_id);
    for (int i = 0; i < num_iterations; i++) {
        unsigned wait_time = wait_times_and_indices[2 * i];
        unsigned index = wait_times_and_indices[2 * i + 1];

        // Give the server some time to generate some key material.
        // NB: There may be "faster" servers on the markets that have key
        // material ready immediately after connecting.  Ours doesn't.
        // So, give it a chance by waiting a wee bit prior to requesting
        // the first key.
        sleep(wait_time);
        gk_request.index = index;

        qkd_get_key_response_t gk_response =
            qkd_get_key_struct(gk_request, &cfg);

        // Print key
        printf("Index: %u\n", gk_response.index);
        qkd_fprintf_status(stdout, gk_response.status);
        printf("\n");
        qkd_fprintf_bytes(stdout, gk_response.key_buffer, 0);

        qkd_get_key_response_dtor(&gk_response);
    }

    // --- CLOSE ---
    qkd_close_request_t cl_request = {};
    uuid_copy(cl_request.key_stream_id, oc_response.key_stream_id);

    // qkd_close() or qkd_close_struct()
    qkd_close_response_t cl_response = qkd_close_struct(cl_request, &cfg);

    // Clean up to keep the likes of Valgrind happy
    qkd_client_context_close_ssl(&cfg);
    qkd_open_connect_response_dtor(&oc_response);
    free(wait_times_and_indices);

    return 0;
}
