#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "examples/ringbuffer.h"

void qkd_rb_init(qkd_rb_t *rb, unsigned size, unsigned key_gen_rate,
                 unsigned chunk_size, unsigned ttf) {
    assert(rb != NULL);
    assert(size != 0);
    assert(key_gen_rate <=
           UINT_MAX / 256); // Paranoia: try to avoid later overflow.
    memset(rb, 0, sizeof(qkd_rb_t));
    rb->init_time = time(NULL);
    rb->key_gen_rate = key_gen_rate;
    rb->chunk_size = chunk_size;
    rb->ttf = ttf;
    rb->size = size;
    rb->data = malloc(size * sizeof(qkd_KSID_index_state_t));
    if (rb->data == NULL) {
        fprintf(stderr, "Error: out of memory\n");
        exit(EXIT_FAILURE);
    }
    for (unsigned i = 0; i < size; i++) {
        rb->data[i] = QKD_INDEX_EMPTY;
    }
    rb->last_fill_time = time(NULL);  // now; filled with ... nothing
    rb->front_idx = rb->back_idx = 1; // zero index has a special meaning
}

void qkd_rb_dtor(qkd_rb_t *rb) {
    assert(rb != NULL);
    free(rb->data);
    memset(rb, 0, sizeof(qkd_rb_t));
}

qkd_KSID_index_state_t qkd_rb_read(qkd_rb_t *rb, unsigned *p_index) {
    assert(rb != NULL);
    assert(p_index != NULL);

    time_t now = time(NULL);
    time_t last_read_time = rb->last_read_time;
    time_t last_fill_time = rb->last_fill_time;
    unsigned last_read_idx = rb->last_read_idx;

    // 1) Try to fill the ringbuffer
    rb->spare_bytes += (now - last_fill_time) * rb->key_gen_rate;
    rb->last_fill_time = now;
    unsigned key_count = rb->key_count;
    for (unsigned i = key_count; i < rb->size; i++) {
        unsigned pos = rb->back_idx % rb->size;
        if (rb->spare_bytes >= rb->chunk_size) {
            rb->spare_bytes -= rb->chunk_size;
            rb->data[pos] = QKD_INDEX_READY;
            rb->key_count++;
            rb->back_idx++;
        } else {
            break;
        }
    }

    // 2) TTF approximation
    //    (Skip this step if this is the first read operation or if TTF=0)
    if (last_read_time != 0 && rb->front_idx != 0 && rb->ttf != 0) {
        double delta_T = last_read_time - rb->init_time;
        delta_T /= rb->front_idx;
        unsigned key_count = rb->key_count;
        // Mark keys with expired TTF as spent
        for (unsigned i = 0; i < key_count; i++) {
            double ttf = last_fill_time + rb->ttf - now +
                         (rb->front_idx + i - last_read_idx) * delta_T;
            unsigned pos = (rb->front_idx + i) % rb->size;
            if (ttf < 0 && rb->data[pos] == QKD_INDEX_READY) {
                rb->data[pos] = QKD_INDEX_EMPTY;
                rb->front_idx++;
                rb->key_count--;
            } else {
                break;
            }
        }
    }

    // 3) Again, try to fill the ringbuffer from spare_bytes
    key_count = rb->key_count;
    for (unsigned i = key_count; i < rb->size; i++) {
        unsigned pos = rb->back_idx % rb->size;
        if (rb->spare_bytes >= rb->chunk_size) {
            rb->spare_bytes -= rb->chunk_size;
            rb->data[pos] = QKD_INDEX_READY;
            rb->key_count++;
            rb->back_idx++;
        } else {
            break;
        }
    }

    // 4) Replace zero index with front_idx
    if (*p_index == 0) {
        *p_index = rb->front_idx;
    }

    // 5) Front access?
    unsigned pos = *p_index % rb->size;
    if (*p_index == rb->front_idx) {
        if (rb->data[pos] == QKD_INDEX_READY) {
            rb->data[pos] = QKD_INDEX_EMPTY;
            rb->front_idx++;
            rb->key_count--;
            // skip spent keys
            while (rb->key_count != 0 &&
                   rb->data[rb->front_idx % rb->size] == QKD_INDEX_SPENT) {
                rb->data[rb->front_idx % rb->size] = QKD_INDEX_EMPTY;
                rb->front_idx++;
                rb->key_count--;
            }
            rb->last_read_time = now;     // update read time ...
            rb->last_read_idx = *p_index; // ... and index
            return QKD_INDEX_READY;
        } else {
            assert(rb->data[pos] == QKD_INDEX_EMPTY);
            return QKD_INDEX_EMPTY;
        }
    }

    // 6) Index out of bounds?
    if (*p_index < rb->front_idx || *p_index >= rb->front_idx + rb->size) {
        return QKD_INDEX_OUT_OF_BOUNDS;
    }

    // 7) Key ready?
    if (rb->data[pos] == QKD_INDEX_READY) {
        rb->data[pos] = QKD_INDEX_SPENT;
        rb->last_read_time = now;     // update read time ...
        rb->last_read_idx = *p_index; // ... and index
        return QKD_INDEX_READY;
    }

    // 8) Empty slot or spent key
    return rb->data[pos];
}
