#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "qkd/cfg_parser.h"

/**
 * Set a value in a key-value pair table.
 * Returns 0 on success.
 * Returns -1 if the key was not found in the table.
 * If the value given is non-NULL then
 * - it is assumed that the caller has malloc()-ed the memory
 * - and that ownership of this memory (including the need to
 *   free() the memory) is transferred to the table.
 */
static int cfg_setval(cfg_table_t table, const char *key, char *val) {
    assert(table != NULL);
    assert(key != NULL);
    cfg_kv_pair_t *element = table;
    while (element->key != NULL) {
        if (strcmp(element->key, key) == 0) {
            free(element->val);
            element->val = val;
            return 0;
        }
        element++;
    }
    return -1; // key not found
}

/**
 * Deallocate memory allocated for the table.
 */
void cfg_table_dtor(cfg_table_t table) {
    assert(table != NULL);
    cfg_kv_pair_t *element = table;
    while (element->key != NULL) {
        free(element->val);
        element->val = NULL;
        element++;
    }
}

int cfg_get_string(cfg_table_t table, const char *key, const char **val) {
    assert(table != NULL);
    assert(key != NULL);
    assert(val != NULL);
    cfg_kv_pair_t *element = table;
    while (element->key != NULL) {
        if (strcmp(element->key, key) == 0) {
            *val = element->val;
            if (element->val == NULL) return 1;
            return 0;
        }
        element++;
    }
    return -1; // key not found
}

int cfg_get_long(cfg_table_t table, const char *key, long *val) {
    assert(table != NULL);
    assert(key != NULL);
    assert(val != NULL);
    cfg_kv_pair_t *element = table;
    while (element->key != NULL) {
        if (strcmp(element->key, key) == 0) {
            if (element->val == NULL) return 1;
            char *invalid_chars = NULL;
            long value = strtol(element->val, &invalid_chars, 10);
            *val = value;
            if (*invalid_chars != 0) return 2;
            return 0;
        }
        element++;
    }
    return -1; // key not found
}

int cfg_from_file(cfg_kv_pair_t *table, const char *filename) {
    if (filename == NULL) return -1;

    FILE *stream = fopen(filename, "r");
    if (stream == NULL) {
        perror("fopen() failed");
        return -1;
    }

    char *line = NULL;
    size_t len = 0;
    for (int lineno = 1;; lineno++) {
        size_t num = getline(&line, &len, stream);
        if (num == -1) { // most likely EOF
            if (errno == EINVAL || errno == ENOMEM) {
                perror("getline() failed");
                fclose(stream);
                return -1;
            }
            free(line);
            break;
        }
        assert(line != NULL);
        if (num == 0 || (num == 1 && line[0] == '\n')) {
            // Skip empty line
            continue;
        }
        if (line[0] == '#') {
            // Skip comment
            continue;
        }
        char *key = NULL;
        char *val = NULL;
        int res = sscanf(line, "%ms = %ms", &key, &val);
        if (res != 2) {
            fprintf(stderr,
                    "Format error on line %d of %s"
                    " ... skipping this line\n",
                    lineno, filename);
        } else {
            if (cfg_setval(table, key, val) == 0) {
                val = NULL; // content moved into table
            } else {
                fprintf(stderr,
                        "Unknown keyword '%s' on line %d of %s, "
                        "skipping entry ...\n",
                        key, lineno, filename);
            }
        }
        free(key);
        key = NULL;
        free(val);
        val = NULL;
    }
    fclose(stream);
    return 0;
}