
#include <stdio.h>
#include "qkd/utils.h"

void qkd_fprintf_bytes(FILE *str, qkd_bytes_t data, size_t group_size) {
    fprintf(str, "Data (%u bytes): ", data.size);
    uint8_t *data_b = data.data;
    size_t i;
    for (i = 0; i < data.size; i++) {
        if (group_size > 0 && i % group_size == 0) fprintf(str, "\n");
        fprintf(str, "%02X ", data_b[i]);
    }
    fprintf(str, "\n");
}

void qkd_fprintf_status(FILE *str, qkd_status_code_t status) {
    const char *s;
    switch (status) {
    case qkd_status_code_successful:
        s = "success";
        break;
    case qkd_status_code_peer_not_connected:
    case qkd_status_code_gk_peer_not_connected:
        s = "peer not connected";
        break;
    case qkd_status_code_gk_insufficient_key_available:
        s = "insufficient key material available";
        break;
    case qkd_status_code_no_qkd_connection_available:
        s = "no QKD connection available";
        break;
    case qkd_status_code_oc_ksid_already_in_use:
        s = "KSID already in use";
        break;
    case qkd_status_code_timeout_error:
        s = "timeout";
        break;
    case qkd_status_code_oc_qos_not_satisfiable:
        s = "QoS not satisfiable";
        break;
    case qkd_status_code_gk_metadata_insufficient_size:
        s = "metadata has insufficient size";
        break;
    case qkd_status_code_out_of_memory:
        s = "out of memory";
        break;
    case qkd_status_code_gk_index_of_out_bounds:
        s = "index out of bounds";
        break;
    case qkd_status_code_gk_key_spent:
        s = "key already spent";
        break;
    case qkd_status_code_oc_no_route:
        s = "no route found";
        break;
    case qkd_status_code_oc_no_discovery_service:
        s = "no application discovery service";
        break;
    case qkd_status_code_oc_protocol_error:
        s = "protocol error";
        break;
    default:
        if (status >= qkd_status_code_vendor_defined) {
            s = "vendor-defined status code";
        } else {
            s = "UNKNOWN STATUS CODE";
        }
    }
    fprintf(str, "Status: %s", s);
}

void qkd_fprintf_hex(FILE *str, unsigned char *data, size_t size,
                     const char *name) {
    fprintf(str, "%s (%zu bytes): ", name, size);
    size_t i;
    for (i = 0; i < size; i++) {
        fprintf(str, "%02X ", data[i]);
    }
    fprintf(str, "\n");
}

void qkd_fprintf_qos(FILE *str, qkd_qos_t qos) {
    fprintf(str, "[%u, %u, %u, %u, %u, %u, \"%s\", \"%s\"]", qos.key_chunk_size,
            qos.max_bps, qos.min_bps, qos.capacity, qos.timeout, qos.ttf,
            qos.oc_metadata_mimetype, qos.gk_metadata_mimetype);
}

void qkd_fprintf_uuid(FILE *str, uuid_t uuid) {
    char uuid_s[UUID_STR_LEN];
    uuid_unparse_lower(uuid, uuid_s);
    fprintf(str, "\"%s\"", uuid_s);
}

void qkd_fprintf_key_buffer(FILE *str, qkd_key_buffer_t key_buffer) {
    if (key_buffer.size < 1) {
        fprintf(str, "[]");
        return;
    }

    uint8_t *data = key_buffer.data;
    fprintf(str, "[%u", data[0]);
    for (size_t i = 1; i < key_buffer.size; i++) fprintf(str, ", %u", data[i]);
    fprintf(str, "]");
}

void qkd_fprintf_oc_request(FILE *str, qkd_open_connect_request_t oc_req) {
    fprintf(str,
            "{\"dest. KME ID\": \"%s\", \"src SAE ID\": \"%s\", \"dest. SAE "
            "ID\": \"%s\", \"qos\": ",
            oc_req.destination_kme_id, oc_req.source_sae_id,
            oc_req.destination_sae_id);
    qkd_fprintf_qos(str, oc_req.qos);
    fprintf(str, ", \"key_stream_id\": ");
    qkd_fprintf_uuid(str, oc_req.key_stream_id);
    fprintf(str, "}");
}

void qkd_fprintf_oc_response(FILE *str, qkd_open_connect_response_t oc_res) {
    fprintf(str, "{\"status\": %d, \"qos\": ", oc_res.status);
    qkd_fprintf_qos(str, oc_res.qos);
    fprintf(str, ", \"key_stream_id\": ");
    qkd_fprintf_uuid(str, oc_res.key_stream_id);
    fprintf(str, "}");
}

void qkd_fprintf_gk_request(FILE *str, qkd_get_key_request_t gk_req) {
    fprintf(str, "{\"key_stream_id\": ");
    qkd_fprintf_uuid(str, gk_req.key_stream_id);
    fprintf(str, ",  \"index\": %u, \"metadata\": %.*s}", gk_req.index,
            (int)gk_req.metadata.size, (char *)gk_req.metadata.data);
}

void qkd_fprintf_gk_response(FILE *str, qkd_get_key_response_t gk_res) {
    fprintf(str,
            "{\"status\": %d, \"index\": %u, \"key_buffer\": ", gk_res.status,
            gk_res.index);
    qkd_fprintf_key_buffer(str, gk_res.key_buffer);
    fprintf(str, ", \"metadata\": %.*s}", (int)gk_res.metadata.size,
            (char *)gk_res.metadata.data);
}

void qkd_fprintf_cl_request(FILE *str, qkd_close_request_t cl_req) {
    fprintf(str, "{\"key_stream_id\": ");
    qkd_fprintf_uuid(str, cl_req.key_stream_id);
    fprintf(str, "}");
}

void qkd_fprintf_cl_response(FILE *str, qkd_close_response_t cl_res) {
    fprintf(str, "{\"status\": %d}", cl_res.status);
}

void qkd_fprintf_header(FILE *str, qkd_header_t header) {
    fprintf(str,
            "{\"version\": [%u, %u, %u], \"protocol_version\": [%u, %u, %u], "
            "\"service\": %u, \"payload_size\": %u}",
            header.version.major, header.version.minor, header.version.micro,
            header.protocol_version.major, header.protocol_version.minor,
            header.protocol_version.micro, header.service, header.payload_size);
}
